#include <ctype.h>
#include <direct.h>
#include <io.h>
#include <process.h>
#include <stdlib.h>
#include <string.h>
#include "x_macros.h"
#include "x_alloc.hpp"
#include "x_cmdlin.hpp"
#include "x_err.hpp"
#include "x_text.hpp"
#include "compress.hpp"
#include "ipx.hpp"
#include "packet.hpp"
#include "port.hpp"
#include "select.hpp"
#include "unique.hpp"

#define SCREEN_FG	LIGHTGRAY
#define SCREEN_BG	BLACK

const char		*title = "LGC IPX Client v1.02 by Adam McKee";
int			nGames;
char			gameName[MAX_GAMES][64];

static const int		stdBase[] = { 0x3f8, 0x2f8, 0x3e8, 0x2e8 };
static const int		stdIRQ[] = { 4, 3, 4, 3 };
static const long		validBPS[] = { 14400, 19200, 28800, 38400, 57600, 115200, 0 };
static char			gamePath[MAX_GAMES][64];
static char			gameEXE[MAX_GAMES][16];
static int			nServerArgs;
static char			*serverArg[16];
static int			nGameArgs;
static char			*gameArg[16];

void			preExit();
int			getToken(char *d, char *s, int *idx);
void			processConfig();
void			buildArgs();
void			launchGame();

void main() {
  int li = 0;
  char line[256], token[256];
  textcolor(SCREEN_FG);
  textbackground(SCREEN_BG);
  clrscr();
  textcolor(WHITE);
  textbackground(RED);
  gotoxy(1,1);
  clreol();
  gotoxy((80 - strlen(title)) / 2, 1);
  cprintf((char*)title);
  gotoxy(1,3);
  atexit(preExit);
  setvbuf(stdout, NULL, _IONBF, 0);
  packet = (byte*)smalloc(MAX_PACKET);
  FILE *f = fopen("lgc.tmp", "rb");
  if (!f) fatalError("lgc.tmp not found.");
  fgets(line, 256, f);
  fclose(f);
  getToken(token, line, &li);
  gameInProgress = atoi(token);
  getToken(token, line, &li);
  initialNodes = atoi(token);
  getToken(token, line, &li);
  nodeAddr = atoi(token);
  while (!getToken(token, line, &li)) {
    if (!stricmp(token, "label")) {
      getToken(token, line, &li);
      strcpy(gameName[nGames], token);
      getToken(gameEXE[nGames], line, &li);
      nGames++;
    } else if (!stricmp(token, "env")) {
      getToken(token, line, &li);
      putenv(token);
    } else if (!stricmp(token, "arg")) {
      getToken(token, line, &li);
      serverArg[nServerArgs] = (char*)smalloc(strlen(token) + 1);
      strcpy(serverArg[nServerArgs], token);
      nServerArgs++;
    }
  }
  processConfig();
  portInit();
  IPXinit();
  if (!nodeAddr)
    if (nGames > 1) selectOptions();
    else setupKnown = 1;
  printf("Using COM%d (%s,0x%x,%d) @ %lu BPS.\n", port + 1, UARTtype ? "16550A" : "8250/16450", base, irq, bps);
  findNodes();
  if (nGames > 1) doSetup();
  buildArgs();
  launchGame();
}

void preExit() {
  int i;
#ifdef DEBUG
  FILE *f = fopen("debug.txt", "at");
  fprintf(f, "----------------------------------------------------------------------------\n");
  fclose(f);
#endif
  for (i = 0; i < 16; i++) if (gameArg[i]) delete gameArg[i];
  if (packet) sfree(packet);
  portDeinit();
  IPXdeInit();
#ifdef DEBUG
  packetDebug();
#endif
#ifdef DEBUG
  compressDebug();
#endif
}

int getToken(char *d, char *s, int *idx) {
  int si, di, slen = strlen(s);
  for (si = *idx; si < slen && isspace(s[si]); si++);
  for (di = 0; si < slen && isgraph(s[si]); si++) d[di++] = s[si];
  d[di] = '\0';
  *idx = si;
  return di == 0;
}

void processConfig() {
  FILE *f = fopen("lgc.cfg", "rt");
  int i, li;
  char line[256], token[64], token2[64];
  if (f == NULL) fatalError("No configuration file!\n");
  while (fgets(line, 256, f) != NULL) {
    li = 0;
    if (!getToken(token, line, &li)) {
      if (token[0] == '#') continue;
      else if (!stricmp(token, "PORT")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	port = atoi(token) - 1;
	if (port < 0 || port > 3) fatalError("Illegal port: %d", 0, port + 1);
	base = stdBase[port];
	irq = stdIRQ[port];
      } else if (!stricmp(token, "BASE")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	sscanf(token, "%x", &base);
      } else if (!stricmp(token, "IRQ")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	irq = atoi(token);
      } else if (!stricmp(token, "BPS")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	bps = atol(token);
	for (i = 0; validBPS[i]; i++) if (bps == validBPS[i]) break;
	if (validBPS[i] == 0) fatalError("Unknown BPS: %s", 0, token);
      } else if (!stricmp(token, "FIFO_SAFETY")) {
        if (getToken(token, line, &li)) continue;
        if (getToken(token, line, &li)) continue;
        fifoSafety = atoi(token);
      } else {
	if (getToken(token2, line, &li)) continue;
	if (getToken(token2, line, &li)) continue;
        for (i = 0; i < nGames; i++)
	  if (!stricmp(token, gameName[i])) {
	    strcpy(gamePath[i], token2);
	    break;
          }
      }
    }
  }
  fclose(f);
}

void buildArgs() {
  int i;
  for (i = 0; i < 16; i++) gameArg[i] = (char*)smalloc(64);
  strcpy(gameArg[nGameArgs++], gameEXE[setup.game]);
  for (i = 0; i < nServerArgs; i++) {
    strcpy(gameArg[nGameArgs++], serverArg[i]);
    sfree(serverArg[i]);
  }
  char buf[64], line[256], token[64];
  sprintf(buf, "%s\\%s", gamePath[setup.game], "lgcargs");
  FILE *f = fopen(buf, "rt");
  if (f) {
    fgets(line, 256, f);
    i = 0;
    while (!getToken(token, line, &i))
      strcpy(gameArg[nGameArgs++], token);
    fclose(f);
  }
  for (i = nGameArgs; i < 16; i++) {
    sfree(gameArg[i]);
    gameArg[i] = NULL;
  }
}

void launchGame() {
  char lgcPath[64], buf[64];
  word lgcDrive, gameDrive, dummy;
  sprintf(buf, "%s\\%s", gamePath[setup.game], gameEXE[setup.game]);
  if (access(buf, 0)) fatalError("path not found: %s", 0, buf);
  getcwd(lgcPath, 64);
  _dos_getdrive(&lgcDrive);
  gameDrive = tolower(gamePath[setup.game][0]) - 'a' + 1;
  _dos_setdrive(gameDrive, &dummy);
  chdir(gamePath[setup.game]);
  printf("\
------------------------------------------------------------------------------\n\n");  
  spawnv(P_WAIT, buf, (char const* const*)gameArg);
  _dos_setdrive(lgcDrive, &dummy);
  chdir(lgcPath);
}
